#include <GL/glut.h>
#include <cmath>
#include <cstdio>
#include <ctime>

static GLdouble angle = .0, cangle = .0;
static bool mouse_left_down = false;
static GLdouble mouse_old_x, mouse_old_y;
static GLdouble current[16];
static GLUquadricObj* quadric;
static GLuint xyzList, objList;

void setmaterial_main(GLfloat r, GLfloat g, GLfloat b)
{
    GLfloat paintcolor[] = { r, g, b, 1.0 };
    GLfloat specularcolor[] = { 1.0, 1.0, 1.0, 1.0 };
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, paintcolor);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specularcolor);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.);
}

void init()
{
    glShadeModel(GL_SMOOTH);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glLoadIdentity();
    glGetDoublev(GL_MODELVIEW_MATRIX, current);
    glLineWidth(5);
    quadric = gluNewQuadric();
    
    xyzList = glGenLists(1);
    objList = glGenLists(1);
    
    glNewList(xyzList, GL_COMPILE);
    // X Y Z
    glBegin(GL_LINES);
    glColor3f(.9, .1, .1);
    glVertex3d(.0, .0, 0);
    glVertex3d(5, .0, 0);
    glColor3f(.1, .9, .1);
    glVertex3d(.0, .0, 0);
    glVertex3d(.0, 5, 0);
    glColor3f(.1, .1, .9);
    glVertex3d(.0, .0, 0);
    glVertex3d(.0, .0, 5);
    glEnd();
    glEndList();
    
    glNewList(objList, GL_COMPILE);
    /* draw a central sphere */
    setmaterial_main(.9, .8, .1);
    glutSolidSphere(0.5, 30, 30);
    /* draw a cube */
    glPushMatrix();
    glTranslated(1.2, .0, .0);
    setmaterial_main(.1, .9, .1);
    glutSolidCube(0.3);
    glPopMatrix();
    /* Learn how to draw a cylinder */
    glPushMatrix();
    setmaterial_main(.9, .1, .1);
    glTranslated(.0, 1.2, .0);
    gluCylinder( quadric, .2, .2, .3, 30, 30 );
    glPopMatrix();
    glEndList();
}

void drawobjects()
{
    GLfloat position[] = { 0., 0., 0., 1. };
    glPushMatrix();
    glTranslated(0.0, 0.0, -5.0);
    glMultMatrixd(current);
    
    glRotated(cangle, 0, 0, 1);
    
    glDisable(GL_LIGHTING);
    glCallList(xyzList);
    glEnable(GL_LIGHTING);
    /* draw a moving light source */
    glPushMatrix();
    glRotated(angle, -1., 1.0, 0.0);
    glTranslated(0.0, 0.0, 1.2);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
    glDisable(GL_LIGHTING);
    glColor3f(1, 1, .8);
    glutSolidSphere(0.1, 30, 30);
    glEnable(GL_LIGHTING);
    glPopMatrix();
    glCallList(objList);
    glPopMatrix();
}

void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    drawobjects();

    glutSwapBuffers();
}

// The GLUT reshape function
void reshape(int w, int h)
{
   glViewport(0, 0, w, h);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   // We set up coordinate system so that aspect ratios are always correct,
   //  and the region from -1..1 in x & y always just fits in the viewport.
   if (w > h)
      glFrustum(-double(w)/h, double(w)/h, -1., 1., 3, 25);
   else
      glFrustum(-1., 1., -double(h)/w, double(h)/w, 3, 25);
   glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}

void timer(int value)
{
    angle += 2.;
    cangle += 0.2;
    if(angle > 360.)
        angle -= 360.;
    glutPostRedisplay();
    
    /* Calculate FPS, why no fps locking in glut? */
    static int time1 = 0, time2, fps = 0;
    time2 = time(0);
    if(time2 - time1 >= 1){
        printf("FPS: %d\n", fps);
        fps = 0, time1 = time2;
    }else
        fps ++;
    glutTimerFunc(16, timer, value);
}

void motion(int x, int y)
{
    if(mouse_left_down){
        GLdouble vx = GLdouble(x - mouse_old_x)/500, vy = GLdouble(y - mouse_old_y)/500;
        GLdouble len = sqrt(vx*vx + vy*vy);
        glPushMatrix();
        glRotated(len*20, vy, vx, 0.);
        glMultMatrixd(current);
        glGetDoublev(GL_MODELVIEW_MATRIX, current);
        glPopMatrix();
        glutPostRedisplay();
    }
}

void zoom(GLdouble x, GLdouble y, GLdouble z)
{
    glPushMatrix();
    glScaled(x, y, z);
    glMultMatrixd(current);
    glGetDoublev(GL_MODELVIEW_MATRIX, current);
    glPopMatrix();
    glutPostRedisplay();
}

void mouse(int button, int state, int x, int y)
{
    if(button == GLUT_LEFT_BUTTON){
        mouse_left_down = state == GLUT_DOWN;
        mouse_old_x = x, mouse_old_y = y;
    }else if(button == 3){
        // Zoom in
        zoom(1.25, 1.25, 1.25);
    }else if(button == 4){
        // Zoom out
        zoom(.8, .8, .8);
    }
}

void keyboard(unsigned char key, int x, int y)
{
    switch(key){
        case 'w':
            zoom(1.25, 1.25, 1.25);
            break;
        case 's':
            zoom(.8, .8, .8);
            break;
    }
}

int main(int argc, char **argv)
{
    printf("Use your mouse to click & drag the world.\n");
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowPosition(100, 100);
    glutInitWindowSize(500, 500);
    glutCreateWindow(argv[0]);
    init();
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutTimerFunc(16, timer, 0);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutKeyboardFunc(keyboard);
    glutMainLoop();
    return 0;
}
